/^Project-Id-Version:/ b
/^Report-Msgid-Bugs-To:/ b
/^POT-Creation-Date:/ b
/^PO-Revision-Date:/ b
/^Last-Translator:/ b
/^Language-Team:/ b
/^Language:/ b
/^MIME-Version:/ b
/^Content-Type:/ b
/^Content-Transfer-Encoding:/ b
/^Plural-Forms:/ b

s,%\([0-9]*\)C,%\1#,g

y,abcdefghijklmnopqrstuvwxyz,αϭϲδεφγχιϊκλμνοπϗρστυβωξϋζ,
y,ABCDEFGHIJKLMNOPQRSTUVWXYZ,ΑϬϲΔΕΦΓΧΙΪΚΛΜΝΟΠϏΡΣΤΥΒΩΞΫΖ,

s,%\([0-9]*\)#,%\1C,g
s,%\([0-9]*\)ϲ,%\1c,g
s,%\([0-9]*\)δ,%\1d,g
s,%\([0-9]*\)λδ,%\1ld,g
s,%\([0-9]*\)λλδ,%\1lld,g

s,%\([0-9\.\*]*\)σ,%\1s,g
s,%\([0-9]*\)λυ,%\1lu,g
s,%\([0-9]*\)υ,%\1u,g
s,%\([0-9]*\)λλυ,%\1llu,g
s,%\([0-9]*\)ξ,%\1x,g
s,%\([0-9]*\)λξ,%\1lx,g
s,%\([0-9]*\)λλξ,%\1llx,g
