/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2010 Free Software Foundation, Inc.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <grub/list.h>
#include <grub/test.h>

int
main (int argc __attribute__ ((unused)),
      char *argv[] __attribute__ ((unused)))
{
  int status = 0;

  grub_test_t test;

  grub_unit_test_init ();
  FOR_LIST_ELEMENTS (test, grub_test_list)
    status = grub_test_run (test) ? : status;

  grub_unit_test_fini ();

  exit (status);
}
