/^Project-Id-Version:/ b
/^Report-Msgid-Bugs-To:/ b
/^POT-Creation-Date:/ b
/^PO-Revision-Date:/ b
/^Last-Translator:/ b
/^Language-Team:/ b
/^Language:/ b
/^MIME-Version:/ b
/^Content-Type:/ b
/^Content-Transfer-Encoding:/ b
/^Plural-Forms:/ b

y,abcdefghijklmnopqrstuvwxyz,абцдефгхийклмнопќрстувўѯѝз,
y,ABCDEFGHIJKLMNOPQRSTUVWXYZ,АБЦДЕФГХИЙКЛМНОПЌРСТУВЎѮЍЗ,

s,%\([0-9]*\)Ц,%\1C,g
s,%\([0-9]*\)ц,%\1c,g
s,%\([0-9]*\)д,%\1d,g
s,%\([0-9]*\)лд,%\1ld,g
s,%\([0-9]*\)ллд,%\1lld,g

s,%\([0-9\.\*]*\)с,%\1s,g
s,%\([0-9]*\)лу,%\1lu,g
s,%\([0-9]*\)у,%\1u,g
s,%\([0-9]*\)ллу,%\1llu,g
s,%\([0-9]*\)ѯ,%\1x,g
s,%\([0-9]*\)лѯ,%\1lx,g
s,%\([0-9]*\)ллѯ,%\1llx,g
